#include <GL/glut.h>
#include <vector>
#include <fstream>

// Some handy macros
#define ABS(x) (((x)>0)?(x):(-x))
#define MAX(x,y) (((x)>(y))?(x):(y))
#define MIN(x,y) (((x)<(y))?(x):(y))

// Some constants
#define MENUHEIGHT (1/8.0f)
#define MENUITEMWIDTH 0.25f
#define MENUITEMCOUNT 4
#define NODESIZE 0.02f

// Think "node"
struct Object {
	float pos[2];
	float vel[2];
};

struct Edge {
	size_t end[2];
};

// A hideous number of global variables
std::vector<Object> objects;
std::vector<Edge> edges;
int ww;
int wh;
float stress;
bool menuSelected[MENUITEMCOUNT] = {true, false, false, false};
bool edgeStarted = false;
size_t selectedNode = 0;
float mx;
float my;
float constant = -0.1f; // Gravitational/repulsive constant
float sep = 0.1f;
bool middleDown = false;

// Returns the index of the nearest node
size_t getNode(float x, float y) {
	size_t minIndex = 0;
	float minDist = -1.0f;
	for (size_t pos = 0; pos < objects.size(); pos++) {
		float delta[2] = {(objects[pos].pos[0] - x), (objects[pos].pos[1] - y)};
		float d2 = delta[0] * delta[0] + delta[1] * delta[1];
		if (d2 < minDist || minDist < 0) {
			minIndex = pos;
			minDist = d2;
		}
	}

	return minIndex;
}

// Does all of the displaying
void display() {
	glClear(GL_COLOR_BUFFER_BIT);
	glPushMatrix();
	glColor3f(1.0f, 1.0f, 0.0f);

	// Draw the edges
	glColor3f(1.0f, 0.0f, 0.0f);
	glBegin(GL_LINES);
	for (std::vector<Edge>::iterator i = edges.begin(); i != edges.end(); i++) {
		if (i->end[0] < objects.size() && i->end[1] < objects.size()) {
			glVertex2f(objects[i->end[0]].pos[0], objects[i->end[0]].pos[1]);
			glVertex2f(objects[i->end[1]].pos[0], objects[i->end[1]].pos[1]);
		}
	}
	glEnd();

	// Draw the nodes
	for (std::vector<Object>::iterator i = objects.begin(); i != objects.end(); i++) {
		if (size_t(getNode(mx, my)) == size_t(i - objects.begin()))
			glColor3f(0.0f, 0.0f, 1.0f);
		else
			glColor3f(1.0f, 1.0f, 0.0f);
		glPushMatrix();
		glTranslatef(i->pos[0], i->pos[1], 0.0f);
		glRectf(-NODESIZE/2, -NODESIZE/2, NODESIZE/2, NODESIZE/2);
		glPopMatrix();
	}

	// Display a half-drawn edge.  It looks neat.
	if (edgeStarted) {
		if (selectedNode < objects.size()) {
			glColor3f(0.0f, 1.0f, 0.0f);
			glBegin(GL_LINES);
			glVertex2f(objects[selectedNode].pos[0], objects[selectedNode].pos[1]);
			glVertex2f(mx, my);
			glEnd();
		}
	}
	
	// Draw the menu boxes and lines
	for (int i = 0; i < MENUITEMCOUNT; i++) {
		float p[2][2] = {{MENUITEMWIDTH * i, MENUITEMWIDTH * (i + 1)}, {1.0f - MENUHEIGHT, 1.0f}};
		glBegin(GL_QUADS);
		if (menuSelected[i])
			glColor3f(1.0f, 1.0f, 1.0f);
		else
			glColor3f(0.5f, 0.5f, 0.5f);
		glVertex2f(p[0][0], p[1][0]);
		glVertex2f(p[0][0], p[1][1]);
		if (menuSelected[i])
			glColor3f(0.7f, 0.7f, 0.7f);
		else
			glColor3f(0.2f, 0.2f, 0.2f);
		glVertex2f(p[0][1], p[1][1]);
		glVertex2f(p[0][1], p[1][0]);
		glEnd();
		glColor3f(0.0f, 0.0f, 0.0f);
		glBegin(GL_LINES);
		glVertex2f(MENUITEMWIDTH * i, 1.0f - MENUHEIGHT);
		glVertex2f(MENUITEMWIDTH * i, 1.0f);
		glEnd();
	}

	// Draw a "node" button.
	glColor3f(1.0f, 1.0f, 0.0f);
	glRectf(MENUITEMWIDTH/2 - MENUITEMWIDTH/3, 1.0f - MENUHEIGHT/2 - MENUHEIGHT/3, MENUITEMWIDTH/2 + MENUITEMWIDTH/3, 1.0f - MENUHEIGHT/2 + MENUHEIGHT/3);

	// Draw an "edge" button.
	glColor3f(0.0f, 1.0f, 0.0f);
	glBegin(GL_LINES);
	glVertex2f(MENUITEMWIDTH*1.2, 1.0f - MENUHEIGHT*0.2f);
	glVertex2f(MENUITEMWIDTH*1.8, 1.0f - MENUHEIGHT*0.8f);
	glEnd();

	// Draw a "clear" button.
	glColor3f(1.0f, 0.0f, 0.0f);
	glBegin(GL_LINES);
	glVertex2f(MENUITEMWIDTH*2.2, 1.0f - MENUHEIGHT*0.2f);
	glVertex2f(MENUITEMWIDTH*2.8, 1.0f - MENUHEIGHT*0.8f);
	glVertex2f(MENUITEMWIDTH*2.8, 1.0f - MENUHEIGHT*0.2f);
	glVertex2f(MENUITEMWIDTH*2.2, 1.0f - MENUHEIGHT*0.8f);
	glEnd();

	// Draw a "play" button.
	glColor3f(0.0f, 0.0f, 1.0f);
	glBegin(GL_TRIANGLES);
	glVertex2f(MENUITEMWIDTH*3.2, 1.0f - MENUHEIGHT*0.2f);
	glVertex2f(MENUITEMWIDTH*3.8, 1.0f - MENUHEIGHT*0.5f);
	glVertex2f(MENUITEMWIDTH*3.2, 1.0f - MENUHEIGHT*0.8f);
	glEnd();
	
	glPopMatrix();
	glutSwapBuffers();
}

// Window resize
void reshape(int w, int h) {
	glViewport(0, 0, ww = w, wh = h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0f, 0.0f, 1.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

// Handle menus and interaction with the graph
void mouse(int button, int state, int x, int y) {
	if (state == GLUT_DOWN) {
		if (float(y) / wh < MENUHEIGHT) { // Handle the toolbar menu
			int item = int((float(x) / ww) / MENUITEMWIDTH);
			switch (item) {
			case 0: // Node
				edgeStarted = false;
				menuSelected[0] = true;
				menuSelected[1] = false;
				break;
			case 1: // Edge
				edgeStarted = false;
				menuSelected[0] = false;
				menuSelected[1] = true;
				break;
			case 2: // Clear
				objects.resize(0);
				edges.resize(0);
				break;
			case 3: // Play
				menuSelected[3] = !menuSelected[3];
				break;
			}
		} else {
			switch (button) {
			case GLUT_LEFT_BUTTON: // Handle nodes/edges
				if (menuSelected[0]) { // Add a node
					Object o = {{float(x) / ww, float(wh - y - 1) / wh}, {0.0f, 0.0f}};
					objects.push_back(o);
				} else if (menuSelected[1]) { // Add an edge
					if (!edgeStarted) {
						edgeStarted = !edgeStarted;
						selectedNode = getNode(float(x) / ww, float(wh - y - 1) / wh);
					}
				}
				break;
			case GLUT_MIDDLE_BUTTON: // Move node
				middleDown = true;
				selectedNode = getNode(mx, my);
				break;
			}
		}
	} else {
		middleDown = false; // Stop moving node
		if (menuSelected[1] && edgeStarted) {
			edgeStarted = !edgeStarted; // Add an edge on mouse release
			Edge e;
			e.end[0] = selectedNode;
			e.end[1] = getNode(float(x) / ww, float(wh - y - 1) / wh);
			if (e.end[1] != e.end[0] && objects.size() > 1)
				edges.push_back(e);
		}
	}
	glutPostRedisplay();
}

void motion(int x, int y) { // Store the mouse position and maybe move nodes
	mx = float(x) / ww;
	my = float(wh - y - 1) / wh;
	if (middleDown && selectedNode < objects.size()) {
		objects[selectedNode].pos[0] = mx;
		objects[selectedNode].pos[1] = my;
		glutPostRedisplay();
	}
}

// Fake physics to make things interesting
void force(Object &a, Object &b, float f[2], bool attract = false){
	f[0] = f[1] = 0.0f;
	float diff[2] = {b.pos[0] - a.pos[0], b.pos[1] - a.pos[1]};
	float c = constant;
	float d2 = (diff[0] * diff[0] + diff[1] * diff[1]);
	float dir[2] = {(diff[0])?(diff[0] / ABS(diff[0])):0.0f, (diff[1])?(diff[1] / ABS(diff[1])):0.0f};

	if (attract && d2 > sep * sep) { // Attract nodes that aren't nearby
		float d = MIN(ABS(d2 - sep * sep), 1.0f);
		for (int i = 0; i < 2; i++)
			f[i] += -c * 2.0f * dir[i] * d;
	} else if (!attract && d2 < sep * sep && d2 > 0.001f) { // Repel nodes that are nearby
		float d = 1.0f - MIN(ABS(d2 - sep * sep), 1.0f);
		for (int i = 0; i < 2; i++)
			f[i] += c * dir[i] * d;
	}
}

// Not much to do to prepare the OpenGL state...
void init() {
	glLineWidth(3.0f);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
}

// Pseudo-accurate animation
void timer(int value) {
	if (menuSelected[3] /* alive */) {
		// Repulsion
 		for (std::vector<Object>::iterator i = objects.begin(); i != objects.end(); i++) {
			float net[2] = {0.0f, 0.0f};
			for (std::vector<Object>::iterator j = objects.begin(); j != objects.end(); j++) {
				if (i != j) {
					float l[2];
					force(*i, *j, l, false);
					net[0] += l[0];
					net[1] += l[1];
				}
			}
			i->vel[0] += net[0];
			i->vel[1] += net[1];
		}
		// Attraction
		for (std::vector<Edge>::iterator i = edges.begin(); i != edges.end(); i++) {
			float net[2] = {0.0f, 0.0f};
			force(objects[i->end[0]], objects[i->end[1]], net, true);
			objects[i->end[0]].vel[0] += net[0];
			objects[i->end[0]].vel[1] += net[1];

			net[0] = net[1] = 0.0f;
			force(objects[i->end[1]], objects[i->end[0]], net, true);
			objects[i->end[1]].vel[0] += net[0];
			objects[i->end[1]].vel[1] += net[1];
		}
		// Motion
		for (std::vector<Object>::iterator i = objects.begin(); i != objects.end(); i++) {
			i->pos[0] += i->vel[0] * 0.01f;
			i->pos[1] += i->vel[1] * 0.01f;

			for (int d = 0; d < 2; d++) {
				if (i->pos[d] > 1.0f) i->pos[d] = 1.0f;
				if (i->pos[d] < 0.0f) i->pos[d] = 0.0f;
			}
		}
		// Friction
		for (std::vector<Object>::iterator i = objects.begin(); i != objects.end(); i++) {
		 	for (int n = 0; n < 2; n++) {
				i->vel[n] *= 0.9f;
			}
		}
	}
	sleep(0);
	glutPostRedisplay();
	glutTimerFunc(50, timer, 0);
}

// For manipulating "constants" and exitting.
void keyboard(unsigned char key, int x, int y) {
	switch (tolower(key)) {
	case '-':
	case '_': constant /= 2.0f; break;
	
	case '=':
	case '+': constant *= 2.0f; break;
	case '0': constant = -0.1f; break;
	
	case 27: // Escape
	case 'q':
		exit(0);
		break;

	case '1': sep = 0.1f; break;
	case '2': sep *= 2.0f; break;
	case '3': sep /= 2.0f; break;
	}
}

// Very very simple graph format
void loadGraph() {
	std::ifstream file("graph.txt");
	size_t nodeCount;
	size_t edgeCount;
	objects.resize(0);
	edges.resize(0);
	file >> nodeCount;
	while (nodeCount--) {
		Object o;
		file >> o.pos[0] >> o.pos[1];
		o.vel[0] = o.vel[1] = 0.0f;
		objects.push_back(o);
	}
	file >> edgeCount;
	while (edgeCount--) {
		Edge e;
		file >> e.end[0] >> e.end[1];
		edges.push_back(e);
	}
	glutPostRedisplay();
}

// Very very simple graph format
void saveGraph() {
	std::ofstream file("graph.txt");
	file << objects.size() << '\n';
	for (std::vector<Object>::iterator i = objects.begin(); i != objects.end(); i++)
		file << i->pos[0] << ' ' << i->pos[1] << '\n';
	file << edges.size() << '\n';
	for (std::vector<Edge>::iterator i = edges.begin(); i != edges.end(); i++)
		file << i->end[0] << ' ' << i->end[1] << '\n';
}

// Handle all of the menus
void menu(int v) {
	switch (v) {
	case 1: { // Remove node
		size_t n = getNode(mx, my);
		if (n < objects.size()) {
			objects.erase(objects.begin() + n);
			for (std::vector<Edge>::iterator i = edges.begin(); i != edges.end(); i++) {
				if (i->end[0] == n || i->end[1] == n) {
					i = edges.erase(i);
					i--;
				} else {
					if (i->end[0] > n)
						i->end[0]--;
					if (i->end[1] > n)
						i->end[1]--;
				}
			}
		}
		} break;
	case 2: // Pause/unpause
		menuSelected[3] = !menuSelected[3];
		break;
	case 3: // Open graph.txt
		loadGraph();
		break;
	case 4: // Save graph.txt
		saveGraph();
		break;
	case 5: // Exit
		exit(0);
		break;
	}
}

// Program entry point
int main(int argc, char *argv[]) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB|GLUT_DOUBLE);
	glutCreateWindow(argv[0]);
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutPassiveMotionFunc(motion);
	glutKeyboardFunc(keyboard);
	init();
	glutInitWindowSize(640, 480);
	glutTimerFunc(0, timer, 0);

	// Make the required menu
	int nodeMenu = glutCreateMenu(menu);
	glutAddMenuEntry("Remove Node", 1);
	int graphMenu = glutCreateMenu(menu);
	glutAddMenuEntry("Pause/Unpause", 2);
	glutAddMenuEntry("Open graph.txt", 3);
	glutAddMenuEntry("Save graph.txt", 4);
	glutAddMenuEntry("Quit", 5);
	glutCreateMenu(menu);
	glutAddSubMenu("Node", nodeMenu);
	glutAddSubMenu("Graph", graphMenu);
	glutAttachMenu(GLUT_RIGHT_BUTTON);
	glutMainLoop();
	return 0;
}
